from uuid import uuid4
from signalwire.blade.handler import trigger, register, unregister, unregister_all
from .constants import CallState, DisconnectReason, ConnectState, CallPlayState, MediaType, RecordType, TapType, CallTapState, CallFaxState, CallSendDigitsState, DetectState, DetectType, PromptState
from .helpers import prepare_prompt_media_list
from .components.dial import Dial
from .components.hangup import Hangup
from .components.answer import Answer
from .components.connect import Connect
from .results.dial_result import DialResult
from .results.hangup_result import HangupResult
from .results.answer_result import AnswerResult
from .results.connect_result import ConnectResult
from .actions.connect_action import ConnectAction
from .components.detect import Detect
from .results.detect_result import DetectResult
from .actions.detect_action import DetectAction
from .components.play import Play
from .results.play_result import PlayResult
from .actions.play_action import PlayAction
from .components.prompt import Prompt
from .results.prompt_result import PromptResult
from .actions.prompt_action import PromptAction
from .components.record import Record
from .results.record_result import RecordResult
from .actions.record_action import RecordAction
from .components.awaiter import Awaiter
from .components.fax_send import FaxSend
from .components.fax_receive import FaxReceive
from .results.fax_result import FaxResult
from .actions.fax_action import FaxAction
from .components.send_digits import SendDigits
from .results.send_digits_result import SendDigitsResult
from .actions.send_digits_action import SendDigitsAction
from .components.tap import Tap
from .results.tap_result import TapResult
from .actions.tap_action import TapAction
from .components.disconnect import Disconnect
from .results.disconnect_result import DisconnectResult

class Call:
  def __init__(self, *, calling, **kwargs):
    self.calling = calling
    self.tag = str(uuid4())
    self.id = kwargs.get('call_id', None)
    self.node_id = kwargs.get('node_id', None)
    self.context = kwargs.get('context', None)
    if 'device' in kwargs:
      self.call_type = kwargs['device'].get('type', None)
      self.from_number = kwargs['device']['params'].get('from_number', None)
      self.to_number = kwargs['device']['params'].get('to_number', None)
      self.timeout = kwargs['device']['params'].get('timeout', None)
    self.prev_state = None
    self.state = kwargs.get('call_state', None)
    self.peer = None
    self.failed = False
    self.busy = False
    self.calling.add_call(self)

  @property
  def device(self):
    device = {
      'type': self.call_type,
      'params': {
        'from_number': self.from_number,
        'to_number': self.to_number
      }
    }
    if self.timeout is not None:
      device['params']['timeout'] = self.timeout
    return device

  @property
  def answered(self):
    return self.state == CallState.ANSWERED

  @property
  def active(self):
    return self.ended == False

  @property
  def ended(self):
    return self.state == CallState.ENDING or self.state == CallState.ENDED

  def on(self, event, callback):
    register(event=self.tag, callback=callback, suffix=event)
    return self

  def off(self, event, callback=None):
    unregister(event=self.tag, callback=callback, suffix=event)
    return self

  async def dial(self):
    component = Dial(self)
    await component.wait_for(CallState.ANSWERED, CallState.ENDING, CallState.ENDED)
    return DialResult(component)

  async def disconnect(self):
    component = Disconnect(self)
    await component.wait_for(ConnectState.FAILED, ConnectState.DISCONNECTED)
    return DisconnectResult(component)

  async def hangup(self, reason: str = 'hangup'):
    component = Hangup(self, reason)
    await component.wait_for(CallState.ENDED)
    return HangupResult(component)

  async def answer(self):
    component = Answer(self)
    await component.wait_for(CallState.ANSWERED, CallState.ENDING, CallState.ENDED)
    return AnswerResult(component)

  async def connect(self, device_list, ringback_list=[]):
    component = Connect(self, device_list, ringback_list)
    await component.wait_for(ConnectState.FAILED, ConnectState.DISCONNECTED, ConnectState.CONNECTED)
    return ConnectResult(component)

  async def connect_async(self, device_list, ringback_list=[]):
    component = Connect(self, device_list, ringback_list)
    await component.execute()
    return ConnectAction(component)

  async def detect(self, detect_type, wait_for_beep=False, timeout=None, initial_timeout=None, end_silence_timeout=None, machine_voice_threshold=None, machine_words_threshold=None, tone=None, digits=None):
    component = Detect(self, detect_type, wait_for_beep, timeout, initial_timeout, end_silence_timeout, machine_voice_threshold, machine_words_threshold, tone, digits)
    await component.wait_for(DetectState.MACHINE, DetectState.HUMAN, DetectState.UNKNOWN, DetectState.CED, DetectState.CNG)
    return DetectResult(component)

  async def detect_async(self, detect_type, wait_for_beep=False, timeout=None, initial_timeout=None, end_silence_timeout=None, machine_voice_threshold=None, machine_words_threshold=None, tone=None, digits=None):
    component = Detect(self, detect_type, wait_for_beep, timeout, initial_timeout, end_silence_timeout, machine_voice_threshold, machine_words_threshold, tone, digits)
    await component.execute()
    return DetectAction(component)

  def detect_answering_machine(self, wait_for_beep=False, timeout=None, initial_timeout=None, end_silence_timeout=None, machine_voice_threshold=None, machine_words_threshold=None):
    return self.detect(DetectType.MACHINE, wait_for_beep=wait_for_beep, timeout=timeout, initial_timeout=initial_timeout, end_silence_timeout=end_silence_timeout, machine_voice_threshold=machine_voice_threshold, machine_words_threshold=machine_words_threshold)

  amd = detect_answering_machine

  def detect_answering_machine_async(self, wait_for_beep=False, timeout=None, initial_timeout=None, end_silence_timeout=None, machine_voice_threshold=None, machine_words_threshold=None):
    return self.detect_async(DetectType.MACHINE, wait_for_beep=wait_for_beep, timeout=timeout, initial_timeout=initial_timeout, end_silence_timeout=end_silence_timeout, machine_voice_threshold=machine_voice_threshold, machine_words_threshold=machine_words_threshold)

  amd_async = detect_answering_machine_async

  def detect_fax(self, tone=None, timeout=None):
    return self.detect(DetectType.FAX, tone=tone, timeout=timeout)

  def detect_fax_async(self, tone=None, timeout=None):
    return self.detect_async(DetectType.FAX, tone=tone, timeout=timeout)

  def detect_digit(self, digits=None, timeout=None):
    return self.detect(DetectType.DIGIT, digits=digits, timeout=timeout)

  def detect_digit_async(self, digits=None, timeout=None):
    return self.detect_async(DetectType.DIGIT, digits=digits, timeout=timeout)

  async def play(self, media_list, volume=0):
    component = Play(self, media_list, volume)
    await component.wait_for(CallPlayState.ERROR, CallPlayState.FINISHED)
    return PlayResult(component)

  async def play_async(self, media_list, volume=0):
    component = Play(self, media_list, volume)
    await component.execute()
    return PlayAction(component)

  def play_audio(self, url, volume=0):
    media_list = [{ 'type': MediaType.AUDIO, 'url': url }]
    return self.play(media_list, volume)

  def play_audio_async(self, url, volume=0):
    media_list = [{ 'type': MediaType.AUDIO, 'url': url }]
    return self.play_async(media_list, volume)

  def play_ringtone(self, name, duration=None, volume=0):
    media = { 'type': MediaType.RINGTONE, 'name': name }
    if duration:
      media['duration'] = float(duration)
    return self.play([ media ], volume)

  def play_ringtone_async(self, name, duration=None, volume=0):
    media = { 'type': MediaType.RINGTONE, 'name': name }
    if duration:
      media['duration'] = float(duration)
    return self.play_async([ media ], volume)

  def play_silence(self, duration):
    media_list = [{ 'type': MediaType.SILENCE, 'duration': float(duration) }]
    return self.play(media_list)

  def play_silence_async(self, duration):
    media_list = [{ 'type': MediaType.SILENCE, 'duration': float(duration) }]
    return self.play_async(media_list)

  def play_tts(self, text, language=None, gender=None, volume=0):
    media = { 'type': MediaType.TTS, 'text': text }
    if language:
      media['language'] = language
    if gender:
      media['gender'] = gender
    return self.play([ media ], volume)

  def play_tts_async(self, text, language=None, gender=None, volume=0):
    media = { 'type': MediaType.TTS, 'text': text }
    if language:
      media['language'] = language
    if gender:
      media['gender'] = gender
    return self.play_async([ media ], volume)

  async def prompt(self, prompt_type, media_list, **kwargs):
    component = Prompt(self, prompt_type, media_list, **kwargs)
    await component.wait_for(PromptState.ERROR, PromptState.NO_INPUT, PromptState.NO_MATCH, PromptState.DIGIT, PromptState.SPEECH)
    return PromptResult(component)

  async def prompt_async(self, prompt_type, media_list, **kwargs):
    component = Prompt(self, prompt_type, media_list, **kwargs)
    await component.execute()
    return PromptAction(component)

  def prompt_audio(self, prompt_type, url, **kwargs):
    media_list = [{ 'type': MediaType.AUDIO, 'url': url }]
    return self.prompt(prompt_type=prompt_type, media_list=media_list, **kwargs)

  def prompt_audio_async(self, prompt_type, url, **kwargs):
    media_list = [{ 'type': MediaType.AUDIO, 'url': url }]
    return self.prompt_async(prompt_type=prompt_type, media_list=media_list, **kwargs)

  def prompt_ringtone(self, prompt_type, name, **kwargs):
    media = prepare_prompt_media_list({ 'name': name }, kwargs)
    media['type'] = MediaType.RINGTONE
    return self.prompt(prompt_type=prompt_type, media_list=[media], **kwargs)

  def prompt_ringtone_async(self, prompt_type, name, **kwargs):
    media = prepare_prompt_media_list({ 'name': name }, kwargs)
    media['type'] = MediaType.RINGTONE
    return self.prompt_async(prompt_type=prompt_type, media_list=[media], **kwargs)

  def prompt_tts(self, prompt_type, text, **kwargs):
    media = prepare_prompt_media_list({ 'text': text }, kwargs)
    media['type'] = MediaType.TTS
    return self.prompt(prompt_type=prompt_type, media_list=[media], **kwargs)

  def prompt_tts_async(self, prompt_type, text, **kwargs):
    media = prepare_prompt_media_list({ 'text': text }, kwargs)
    media['type'] = MediaType.TTS
    return self.prompt_async(prompt_type=prompt_type, media_list=[media], **kwargs)

  async def record(self, record_type=RecordType.AUDIO, beep=None, record_format=None, stereo=None, direction=None, initial_timeout=None, end_silence_timeout=None, terminators=None):
    component = Record(self, record_type, beep, record_format, stereo, direction, initial_timeout, end_silence_timeout, terminators)
    await component.wait_for(CallPlayState.ERROR, CallPlayState.FINISHED)
    return RecordResult(component)

  async def record_async(self, record_type=RecordType.AUDIO, beep=None, record_format=None, stereo=None, direction=None, initial_timeout=None, end_silence_timeout=None, terminators=None):
    component = Record(self, record_type, beep, record_format, stereo, direction, initial_timeout, end_silence_timeout, terminators)
    result = await component.execute()
    if result and 'url' in result:
      component.url = result['url']
    return RecordAction(component)

  async def send_digits(self, digits):
    component = SendDigits(self, digits)
    await component.wait_for(CallSendDigitsState.FINISHED)
    return SendDigitsResult(component)

  async def send_digits_async(self, digits):
    component = SendDigits(self, digits)
    await component.execute()
    return SendDigitsAction(component)

  async def wait_for(self, events=[CallState.ENDED]):
    state_index = CallState.ALL.index(self.state)
    for event in events:
      if CallState.ALL.index(event) <= state_index:
        return True
    component = Awaiter(self)
    await component.wait_for(*events)
    return component.successful

  def wait_for_ringing(self):
    return self.wait_for(events=[CallState.RINGING])

  def wait_for_answered(self):
    return self.wait_for(events=[CallState.ANSWERED])

  def wait_for_ending(self):
    return self.wait_for(events=[CallState.ENDING])

  def wait_for_ended(self):
    return self.wait_for(events=[CallState.ENDED])

  async def fax_receive(self):
    component = FaxReceive(self)
    await component.wait_for(CallFaxState.ERROR, CallFaxState.FINISHED)
    return FaxResult(component)

  async def fax_receive_async(self):
    component = FaxReceive(self)
    await component.execute()
    return FaxAction(component)

  async def fax_send(self, url, identity=None, header=None):
    component = FaxSend(self, document=url, identity=identity, header=header)
    await component.wait_for(CallFaxState.ERROR, CallFaxState.FINISHED)
    return FaxResult(component)

  async def fax_send_async(self, url, identity=None, header=None):
    component = FaxSend(self, document=url, identity=identity, header=header)
    await component.execute()
    return FaxAction(component)

  async def tap(self, audio_direction, target_type, target_addr=None, target_port=None, target_ptime=None, target_uri=None, rate=None, codec=None):
    component = Tap(self, audio_direction, target_type, target_addr, target_port, target_ptime, target_uri, rate, codec)
    await component.wait_for(CallTapState.FINISHED)
    return TapResult(component)

  async def tap_async(self, audio_direction, target_type, target_addr=None, target_port=None, target_ptime=None, target_uri=None, rate=None, codec=None):
    component = Tap(self, audio_direction, target_type, target_addr, target_port, target_ptime, target_uri, rate, codec)
    await component.execute()
    return TapAction(component)

  def _state_changed(self, params):
    self.prev_state = self.state
    self.state = params['call_state']
    trigger(self.tag, params, suffix='stateChange')
    trigger(self.tag, params, suffix=self.state)
    if self.state == CallState.ENDED:
      check_id = self.id if self.id else self.tag
      trigger(check_id, params, suffix=CallState.ENDED) # terminate components
      unregister_all(self.tag) # unregister all external handlers
      end_reason = params.get('end_reason', '')
      self.failed = end_reason == DisconnectReason.ERROR
      self.busy = end_reason == DisconnectReason.BUSY
      self.calling.remove_call(self)



from abc import ABC
from ..components import BaseComponent

class BaseResult(ABC):
  def __init__(self, component: BaseComponent):
    self.component = component

  @property
  def successful(self):
    return self.component.successful

  @property
  def event(self):
    return self.component.event



from . import BaseResult

class TapResult(BaseResult):

  @property
  def tap(self):
    return self.component.tap

  @property
  def source_device(self):
    return self.component.source_device

  @property
  def destination_device(self):
    return self.component.device



from abc import ABC, abstractmethod, abstractproperty
import logging
from signalwire.blade.handler import register
from .helpers import receive_contexts

class BaseRelay(ABC):
  def __init__(self, client):
    self.client = client

  @abstractproperty
  def service(self):
    pass

  @abstractmethod
  def notification_handler(self, notification):
    pass

  def ctx_receive_unique(self, context):
    return f"{self.service}.ctx_receive.{context}"

  def ctx_state_unique(self, context):
    return f"{self.service}.ctx_state.{context}"

  async def receive(self, contexts, handler):
    try:
      await receive_contexts(self.client, contexts)
      for context in contexts:
        register(event=self.client.protocol, callback=handler, suffix=self.ctx_receive_unique(context))
    except Exception as error:
      logging.error('receive error: {0}'.format(str(error)))

  async def state_change(self, contexts, handler):
    try:
      await receive_contexts(self.client, contexts)
      for context in contexts:
        register(event=self.client.protocol, callback=handler, suffix=self.ctx_state_unique(context))
    except Exception as error:
      logging.error('state_change error: {0}'.format(str(error)))



import asyncio
import logging
from abc import ABC, abstractmethod, abstractproperty
from uuid import uuid4
from signalwire.blade.handler import register, unregister
from signalwire.blade.messages.execute import Execute
from ..constants import CallState
from ...event import Event

class BaseComponent(ABC):
  def __init__(self, call):
    self.call = call
    self.control_id = str(uuid4())
    self.state = ''
    self.completed = False
    self.successful = False
    self.event = None
    self._future = None
    self._execute_result = None
    self._events_to_await = []

  @abstractproperty
  def event_type(self):
    pass

  @abstractproperty
  def method(self):
    pass

  @abstractproperty
  def payload(self):
    pass

  @abstractmethod
  def notification_handler(self, params):
    pass

  def register(self):
    register(event=self.event_type, callback=self.notification_handler, suffix=self.control_id)
    check_id = self.call.id if self.call.id else self.call.tag
    register(event=check_id, callback=self.terminate, suffix=CallState.ENDED)

  def unregister(self):
    unregister(event=self.event_type, callback=self.notification_handler, suffix=self.control_id)
    if self.call.id:
      unregister(event=self.call.id, callback=self.terminate, suffix=CallState.ENDED)
    unregister(event=self.call.tag, callback=self.terminate, suffix=CallState.ENDED)

  async def execute(self):
    if self.call.ended == True:
      return self.terminate()
    self.register()
    if self.method is None:
      return
    msg = Execute({
      'protocol': self.call.calling.client.protocol,
      'method': self.method,
      'params': self.payload
    })
    try:
      response = await self.call.calling.client.execute(msg)
      self._execute_result = response.get('result', {})
      return self._execute_result
    except Exception as error:
      logging.error('Relay command failed: {0}'.format(str(error)))
      self.terminate()

  async def wait_for(self, *events):
    self._events_to_await = events
    self._future = asyncio.get_event_loop().create_future()
    await self.execute()
    try:
      await self._future
    except asyncio.CancelledError:
      pass

  def terminate(self, params={}):
    self.unregister()
    self.completed = True
    self.successful = False
    self.state = 'failed'
    if 'call_state' in params:
      self.event = Event(params['call_state'], params)
    if self.has_future():
      self._future.cancel()

  def has_future(self):
    return isinstance(self._future, asyncio.Future)



import logging
from .constants import BladeMethod

def handle_inbound_message(client, message):
  if message.method == BladeMethod.NETCAST:
    pass
  elif message.method == BladeMethod.BROADCAST:
    _blade_broadcast(client, message.params)
  elif message.method == BladeMethod.DISCONNECT:
    client._idle = True

def _blade_broadcast(client, params):
  if client.protocol != params['protocol']:
    logging.warn('Client protocol mismatch.')
    return

  if params['event'] == 'queuing.relay.events':
    # FIXME: at the moment all these events are for calling. In the future we'll change the case
    client.calling.notification_handler(params['params'])
  elif params['event'] == 'queuing.relay.tasks':
    client.tasking.notification_handler(params['params'])
  elif params['event'] == 'queuing.relay.messaging':
    client.messaging.notification_handler(params['params'])



from unittest import TestCase

class TestFaxResponse(TestCase):
  def test_returns_laml(self):
    from signalwire.fax_response import FaxResponse
    r = FaxResponse()
    r.receive(action = '/foo/bar')
    self.assertEqual(str(r), '<?xml version="1.0" encoding="UTF-8"?><Response><Receive action="/foo/bar" /></Response>')

  def test_reject_laml(self):
    from signalwire.fax_response import FaxResponse
    r = FaxResponse()
    obj = r.reject()
    self.assertEqual(str(r), '<?xml version="1.0" encoding="UTF-8"?><Response><Reject /></Response>')





from unittest import TestCase
import os

class TestConfigurable(TestCase):
  def test_api_url(self):
    from signalwire.rest import Client as signalwire_client
    client = signalwire_client('account', 'token', signalwire_space_url = 'myname.signalwire.com')

    self.assertEqual(client.api.base_url, 'https://myname.signalwire.com')

  def test_api_url_scheme(self):
    from signalwire.rest import Client as signalwire_client
    client = signalwire_client('account', 'token', signalwire_space_url = 'https://myname.signalwire.com')

    self.assertEqual(client.api.base_url, 'https://myname.signalwire.com')
  
  def test_api_url_environnment(self):
    from signalwire.rest import Client as signalwire_client
    os.environ['SIGNALWIRE_SPACE_URL'] = 'myname.signalwire.com'
    client = signalwire_client('account', 'token')

    self.assertEqual(client.api.base_url, 'https://myname.signalwire.com')



from . import BaseComponent
from ..constants import Notification, CallFaxState
from ...event import Event
from .decorators import stoppable

@stoppable
class BaseFax(BaseComponent):

  def __init__(self, call):
    super().__init__(call)
    self.direction = None
    self.identity = None
    self.remote_identity = None
    self.document = None
    self.pages = None

  @property
  def event_type(self):
    return Notification.FAX

  def notification_handler(self, params):
    try:
      fax = params['fax']
      self.state = fax['type']
    except KeyError:
      return

    self.completed = self.state != CallFaxState.PAGE
    if self.completed:
      self.unregister()
      self.event = Event(self.state, fax)
      self.successful = fax['params'].get('success', False)
      if self.successful:
        self.direction = fax['params'].get('direction', None)
        self.identity = fax['params'].get('identity', None)
        self.remote_identity = fax['params'].get('remote_identity', None)
        self.document = fax['params'].get('document', None)
        self.pages = fax['params'].get('pages', None)
      if self.has_future():
        self._future.set_result(True)



from . import BaseComponent
from ..constants import Method, Notification, CallSendDigitsState
from ...event import Event

class SendDigits(BaseComponent):

  def __init__(self, call, digits):
    super().__init__(call)
    self.digits = digits

  @property
  def event_type(self):
    return Notification.SEND_DIGITS

  @property
  def method(self):
    return Method.SEND_DIGITS

  @property
  def payload(self):
    return {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id,
      'digits': self.digits
    }

  def notification_handler(self, params):
    self.state = params.get('state', None)
    if self.state is None:
      return

    self.completed = self.state == CallSendDigitsState.FINISHED
    self.successful = self.completed
    self.unregister()
    self.event = Event(self.state, params)
    if self.has_future():
      self._future.set_result(True)



from twilio.twiml.fax_response import FaxResponse as TwilioFaxResponse
from twilio.twiml import TwiML

class FaxResponse(TwilioFaxResponse):
  def reject(self, **kwargs):
    """
    Create a <Reject> element
    :param kwargs: additional attributes
    :returns: <Reject> element
    """
    return self.nest(Reject(**kwargs))

class Reject(TwiML):
  """ <Reject> TwiML Verb """

  def __init__(self, **kwargs):
    super(Reject, self).__init__(**kwargs)
    self.name = 'Reject'



import json
from uuid import uuid4

class Message:
  def __init__(self, **kwargs):
    self.jsonrpc = '2.0'
    self.id = kwargs.pop('id', str(uuid4()))
    if 'method' in kwargs:
      self.method = kwargs.pop('method')
    if 'params' in kwargs:
      self.params = kwargs.pop('params')
    if 'error' in kwargs:
      self.error = kwargs.pop('error')
    if 'result' in kwargs:
      self.result = kwargs.pop('result')

  def to_json(self, **kwargs):
    return json.dumps(self.__dict__, separators=(',', ':'), **kwargs)

  @classmethod
  def from_json(cls, json_str):
    json_dict = json.loads(json_str)
    return cls(**json_dict)



from . import BaseResult

class PromptResult(BaseResult):

  @property
  def prompt_type(self):
    return self.component.prompt_type

  @property
  def result(self):
    return self.component.input

  @property
  def terminator(self):
    return self.component.terminator

  @property
  def confidence(self):
    return self.component.confidence

class PromptVolumeResult:
  def __init__(self, successful):
    self.successful = successful



import logging
from signalwire.blade.handler import trigger
from signalwire.relay import BaseRelay

class Tasking(BaseRelay):

  @property
  def service(self):
    return 'tasking'

  def notification_handler(self, notification):
    context = notification['context']
    logging.info(f'Receive task in context: {context}')
    trigger(self.client.protocol, notification['message'], suffix=self.ctx_receive_unique(context))



try:
  from urllib.parse import urlparse, ParseResult
except ImportError:
  from urlparse import urlparse, ParseResult
from twilio.rest import Client as TwilioClient
from twilio.rest.api import Api as TwilioApi
from twilio.base.exceptions import TwilioRestException
from twilio.base import deserialize

from twilio.rest.api.v2010.account.application import ApplicationInstance
from twilio.rest.api.v2010.account import AccountInstance
from twilio.rest.api.v2010.account.call import CallInstance
from twilio.rest.api.v2010.account.recording import RecordingInstance
from twilio.rest.api.v2010.account.transcription import TranscriptionInstance
from twilio.rest.api.v2010.account.message import MessageInstance
from twilio.rest.api.v2010.account.available_phone_number.local import LocalInstance
from twilio.rest.api.v2010.account.available_phone_number.toll_free import TollFreeInstance
from twilio.rest.api.v2010.account.incoming_phone_number import IncomingPhoneNumberInstance

from twilio.rest.fax import Fax as TwilioFax
from twilio.rest.fax.v1 import V1 as TwilioV1

import sys
from six import u
import os

def patched_str(self):
    """ Try to pretty-print the exception, if this is going on screen. """

    def red(words):
      return u("\033[31m\033[49m%s\033[0m") % words

    def white(words):
      return u("\033[37m\033[49m%s\033[0m") % words

    def blue(words):
      return u("\033[34m\033[49m%s\033[0m") % words

    def teal(words):
      return u("\033[36m\033[49m%s\033[0m") % words

    def get_uri(code):
       return "https://www.signalwire.com/docs/errors/{0}".format(code)

    # If it makes sense to print a human readable error message, try to
    # do it. The one problem is that someone might catch this error and
    # try to display the message from it to an end user.
    if hasattr(sys.stderr, 'isatty') and sys.stderr.isatty():
      msg = (
        "\n{red_error} {request_was}\n\n{http_line}"
        "\n\n{sw_returned}\n\n{message}\n".format(
          red_error=red("HTTP Error"),
          request_was=white("Your request was:"),
          http_line=teal("%s %s" % (self.method, self.uri)),
          sw_returned=white(
            "Signalwire returned the following information:"),
            message=blue(str(self.msg))
          ))
      if self.code:
        msg = "".join([msg, "\n{more_info}\n\n{uri}\n\n".format(
          more_info=white("More information may be available here:"),
          uri=blue(get_uri(self.code))),
        ])
      return msg
    else:
      return "HTTP {0} error: {1}".format(self.status, self.msg)

def patched_applicationinstance_init(self, version, payload, account_sid, sid=None):
    """
    Initialize the ApplicationInstance

    :returns: twilio.rest.api.v2010.account.application.ApplicationInstance
    :rtype: twilio.rest.api.v2010.account.application.ApplicationInstance
    """
    super(ApplicationInstance, self).__init__(version)

    # Marshaled Properties
    self._properties = {
        'account_sid': payload['account_sid'],
        'api_version': payload['api_version'],
        'date_created': deserialize.rfc2822_datetime(payload['date_created']),
        'date_updated': deserialize.rfc2822_datetime(payload['date_updated']),
        'friendly_name': payload['friendly_name'],
        'message_status_callback': payload.get('message_status_callback', ''), #missing
        'sid': payload['sid'],
        'sms_fallback_method': payload['sms_fallback_method'],
        'sms_fallback_url': payload['sms_fallback_url'],
        'sms_method': payload['sms_method'],
        'sms_status_callback': payload['sms_status_callback'],
        'sms_url': payload['sms_url'],
        'status_callback': payload['status_callback'],
        'status_callback_method': payload['status_callback_method'],
        'uri': payload['uri'],
        'voice_caller_id_lookup': payload['voice_caller_id_lookup'],
        'voice_fallback_method': payload['voice_fallback_method'],
        'voice_fallback_url': payload['voice_fallback_url'],
        'voice_method': payload['voice_method'],
        'voice_url': payload['voice_url'],
    }

    # Context
    self._context = None
    self._solution = {'account_sid': account_sid, 'sid': sid or self._properties['sid'], }

def patched_accountinstance_init(self, version, payload, sid=None):
        """
        Initialize the AccountInstance
        :returns: twilio.rest.api.v2010.account.AccountInstance
        :rtype: twilio.rest.api.v2010.account.AccountInstance
        """
        super(AccountInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'auth_token': payload['auth_token'],
            'date_created': deserialize.rfc2822_datetime(payload['date_created']),
            'date_updated': deserialize.rfc2822_datetime(payload['date_updated']),
            'friendly_name': payload['friendly_name'],
            'owner_account_sid': payload.get('owner_account_sid', ''),
            'sid': payload['sid'],
            'status': payload['status'],
            'subresource_uris': payload['subresource_uris'],
            'type': payload['type'],
            'uri': payload['uri'],
        }

        # Context
        self._context = None
        self._solution = {'sid': sid or self._properties['sid'], }


def patched_localinstance_init(self, version, payload, account_sid, country_code):
    """
    Initialize the LocalInstance
    :returns: twilio.rest.api.v2010.account.available_phone_number.local.LocalInstance
    :rtype: twilio.rest.api.v2010.account.available_phone_number.local.LocalInstance
    """
    super(LocalInstance, self).__init__(version)

    # Marshaled Properties
    self._properties = {
        'friendly_name': payload['friendly_name'],
        'phone_number': payload['phone_number'],
        'lata': payload['lata'],
        'locality': payload.get('locality', ''), #missing
        'rate_center': payload['rate_center'],
        'latitude': deserialize.decimal(payload['latitude']),
        'longitude': deserialize.decimal(payload['longitude']),
        'region': payload['region'],
        'postal_code': payload['postal_code'],
        'iso_country': payload['iso_country'],
        'beta': payload['beta'],
        'capabilities': payload['capabilities'],
    }

    # Context
    self._context = None
    self._solution = {'account_sid': account_sid, 'country_code': country_code, }

def patched_incomingphonenumberinstance_init(self, version, payload, account_sid, sid=None):
    """
    Initialize the IncomingPhoneNumberInstance

    :returns: twilio.rest.api.v2010.account.incoming_phone_number.IncomingPhoneNumberInstance
    :rtype: twilio.rest.api.v2010.account.incoming_phone_number.IncomingPhoneNumberInstance
    """
    super(IncomingPhoneNumberInstance, self).__init__(version)

    # Marshaled Properties
    self._properties = {
        'account_sid': payload.get('account_sid', ''), #missing
        'address_sid': payload.get('address_sid', ''), #missing
        'address_requirements': payload.get('address_requirements', ''), #missing
        'api_version': payload['api_version'],
        'beta': payload['beta'],
        'capabilities': payload['capabilities'],
        'date_created': deserialize.rfc2822_datetime(payload['date_created']),
        'date_updated': deserialize.rfc2822_datetime(payload['date_updated']),
        'friendly_name': payload['friendly_name'],
        'identity_sid': payload.get('identity_sid', ''), #missing,
        'phone_number': payload['phone_number'],
        'origin': payload.get('origin', ''), #missing,
        'sid': payload['sid'],
        'sms_application_sid': payload['sms_application_sid'],
        'sms_fallback_method': payload['sms_fallback_method'],
        'sms_fallback_url': payload['sms_fallback_url'],
        'sms_method': payload['sms_method'],
        'sms_url': payload['sms_url'],
        'status_callback': payload['status_callback'],
        'status_callback_method': payload['status_callback_method'],
        'trunk_sid': payload.get('trunk_sid', ''), #missing,
        'uri': payload['uri'],
        'voice_application_sid': payload['voice_application_sid'],
        'voice_caller_id_lookup': payload['voice_caller_id_lookup'],
        'voice_fallback_method': payload['voice_fallback_method'],
        'voice_fallback_url': payload['voice_fallback_url'],
        'voice_method': payload['voice_method'],
        'voice_url': payload['voice_url'],
        'emergency_status': payload.get('emergency_status', ''), #missing,
        'emergency_address_sid': payload.get('emergency_address_sid', ''), #missing,
        }

    # Context
    self._context = None
    self._solution = {'account_sid': account_sid, 'sid': sid or self._properties['sid'], }

def patched_tollfreeinstance_init(self, version, payload, account_sid, country_code):
    """
    Initialize the TollFreeInstance

    :returns: twilio.rest.api.v2010.account.available_phone_number.toll_free.TollFreeInstance
    :rtype: twilio.rest.api.v2010.account.available_phone_number.toll_free.TollFreeInstance
    """
    super(TollFreeInstance, self).__init__(version)

    # Marshaled Properties
    self._properties = {
        'friendly_name': payload['friendly_name'],
        'phone_number': payload['phone_number'],
        'lata': payload['lata'],
        'locality': payload.get('locality', ''), #missing
        'rate_center': payload['rate_center'],
        'latitude': deserialize.decimal(payload['latitude']),
        'longitude': deserialize.decimal(payload['longitude']),
        'region': payload['region'],
        'postal_code': payload['postal_code'],
        'iso_country': payload['iso_country'],
        'beta': payload['beta'],
        'capabilities': payload['capabilities'],
    }

    # Context
    self._context = None
    self._solution = {'account_sid': account_sid, 'country_code': country_code, }

def patched_recordinginstance_init(self, version, payload, account_sid, sid=None):
    """
    Initialize the RecordingInstance
    :returns: twilio.rest.api.v2010.account.call.recording.RecordingInstance
    :rtype: twilio.rest.api.v2010.account.call.recording.RecordingInstance
    """
    super(RecordingInstance, self).__init__(version)

    # Marshaled Properties
    self._properties = {
        'account_sid': payload['account_sid'],
        'api_version': payload['api_version'],
        'call_sid': payload['call_sid'],
        'conference_sid': payload['conference_sid'],
        'date_created': deserialize.rfc2822_datetime(payload['date_created']),
        'date_updated': deserialize.rfc2822_datetime(payload['date_updated']),
        'start_time': deserialize.rfc2822_datetime(payload['start_time']),
        'duration': payload['duration'],
        'sid': payload['sid'],
        'price': deserialize.decimal(payload['price']),
        'uri': payload['uri'],
        'encryption_details': payload.get('encryption_details', ''), #missing
        # 'encryption_details': payload['encryption_details'],
        'price_unit': payload['price_unit'],
        'status': payload['status'],
        'channels': deserialize.integer(payload.get('channels', 1)), #missing
        # 'channels': deserialize.integer(payload['channels']),
        'source': payload['source'],
        'error_code': deserialize.integer(payload['error_code']),
    }

    # Context
    self._context = None
    self._solution = {'account_sid': account_sid, 'sid': sid or self._properties['sid'], }

def patched_transcriptioninstance_init(self, version, payload, account_sid, sid=None):
      """
      Initialize the TranscriptionInstance
      :returns: twilio.rest.api.v2010.account.transcription.TranscriptionInstance
      :rtype: twilio.rest.api.v2010.account.transcription.TranscriptionInstance
      """
      super(TranscriptionInstance, self).__init__(version)

      # Marshaled Properties
      self._properties = {
          'account_sid': payload['account_sid'],
          'api_version': payload['api_version'],
          'date_created': deserialize.rfc2822_datetime(payload['date_created']),
          'date_updated': deserialize.rfc2822_datetime(payload['date_updated']),
          'duration': payload['duration'],
          'price': deserialize.decimal(payload['price']),
          'price_unit': payload['price_unit'],
          'recording_sid': payload['recording_sid'],
          'sid': payload['sid'],
          'status': payload['status'],
          'transcription_text': payload['transcription_text'],
          'type': payload.get('type', ''), #missing parameter
          'uri': payload['uri'],
      }

      # Context
      self._context = None
      self._solution = {'account_sid': account_sid, 'sid': sid or self._properties['sid'], }

def patched_fax_init(self, twilio):
  """
  Initialize the Fax Domain
  :returns: Domain for Fax
  :rtype: twilio.rest.fax.Fax
  """
  super(TwilioFax, self).__init__(twilio)

  self.base_url = ''
  self.account_sid = twilio.account_sid

  # Versions
  self._v1 = None

def patched_fax_v1_init(self, domain):
  """
  Initialize the V1 version of Fax
  :returns: V1 version of Fax
  :rtype: twilio.rest.fax.v1.V1.V1
  """
  super(TwilioV1, self).__init__(domain)
  self.version = "2010-04-01/Accounts/" + domain.account_sid
  self._faxes = None


class Client(TwilioClient):
  def __init__(self, *args, **kwargs):
    if 'signalwire_space_url' in kwargs:
      signalwire_space_url = kwargs.pop('signalwire_space_url', "api.signalwire.com")
    else:
      signalwire_space_url = os.environ['SIGNALWIRE_SPACE_URL']

    p = urlparse(signalwire_space_url, 'http')
    netloc = p.netloc or p.path
    path = p.path if p.netloc else ''
    p = ParseResult('https', netloc, path, *p[3:])

    super(Client, self).__init__(*args, **kwargs)
    self._api = TwilioApi(self)
    self._api.base_url = p.geturl()

    TwilioFax.__init__ = patched_fax_init
    TwilioV1.__init__ = patched_fax_v1_init

    self._fax = TwilioFax(self)
    self._fax.base_url = p.geturl()

    TwilioRestException.__str__ = patched_str
    AccountInstance.__init__ = patched_accountinstance_init
    LocalInstance.__init__ = patched_localinstance_init
    TollFreeInstance.__init__ = patched_tollfreeinstance_init
    ApplicationInstance.__init__ = patched_applicationinstance_init
    IncomingPhoneNumberInstance.__init__ = patched_incomingphonenumberinstance_init
    RecordingInstance.__init__ = patched_recordinginstance_init
    TranscriptionInstance.__init__ = patched_transcriptioninstance_init



from signalwire.blade.messages.message import Message

class Ping(Message):

  def __init__(self, timestamp=None):
    self.method = 'blade.ping'
    params = { 'timestamp': timestamp } if timestamp else {}
    super().__init__(params=params)



from .helpers import safe_invoke_callback

GLOBAL = 'GLOBAL'
_queue = {}

def register(*, event, callback, suffix=GLOBAL):
  global _queue

  event = build_event_name(event, suffix)
  if event not in _queue:
    _queue[event] = []
  _queue[event].append(callback)

def register_once(*, event, callback, suffix=GLOBAL):
  global _queue

  def cb(*args):
    unregister(event=event, callback=cb, suffix=suffix)
    callback(*args)

  register(event=event, callback=cb, suffix=suffix)

def unregister(*, event, callback=None, suffix=GLOBAL):
  global _queue

  if is_queued(event, suffix) is False:
    return False
  event = build_event_name(event, suffix)
  if callback is None:
    _queue[event] = []
  else:
    for index, handler in enumerate(_queue[event]):
      if callback == handler:
        del _queue[event][index]

  if (_queue[event]) == 0:
    del _queue[event]

  return True

def unregister_all(event):
  global _queue

  target = build_event_name(event, '')
  for event in list(_queue.keys()):
    if event.find(target) == 0:
      del _queue[event]

def trigger(event, *args, suffix=GLOBAL, **kwargs):
  global _queue

  if is_queued(event, suffix) is False:
    return False
  event = build_event_name(event, suffix)
  for callback in _queue[event]:
    safe_invoke_callback(callback, *args, **kwargs)
  return True

def is_queued(event, suffix=GLOBAL):
  global _queue

  event = build_event_name(event, suffix)
  return event in _queue and len(_queue[event]) > 0

def queue_size(event, suffix=GLOBAL):
  global _queue

  if is_queued(event, suffix) is False:
    return 0
  event = build_event_name(event, suffix)
  return len(_queue[event])

def clear():
  global _queue

  _queue = {}

def build_event_name(event, suffix):
  return "{0}|{1}".format(event.strip(), suffix.strip())



from aiohttp import WSMsgType, ClientSession, ClientWebSocketResponse
import logging
import re
from signalwire.blade.messages.message import Message

class Connection:
  def __init__(self, client, session=ClientSession):
    self._session = session()
    self.client = client
    self.host = self._checkHost(client.host)
    self.ws = None
    self._requests = {}

  @property
  def connected(self):
    return self.ws is not None and self.ws.closed == False

  def _checkHost(self, host):
    # TODO: move to a helper
    protocol = '' if re.match(r"^(ws|wss):\/\/", host) else 'wss://'
    return protocol + host

  async def send(self, message):
    logging.debug('SEND: \n' + message.to_json(indent=2))
    await self.ws.send_str(message.to_json())

  async def connect(self):
    logging.debug('Connecting to: {0}'.format(self.host))
    self.ws = await self._session.ws_connect(self.host)

  async def read(self):
    async for msg in self.ws:
      logging.debug('RECV: \n' + msg.data)
      if msg.type == WSMsgType.TEXT:
        self.client.message_handler(Message.from_json(msg.data))
      elif msg.type == WSMsgType.CLOSED:
        logging.info('WebSocket Closed!')
        break
      elif msg.type == WSMsgType.ERROR:
        logging.info('WebSocket Error!')
        break

  async def close(self):
    if self.connected:
      await self.ws.close()



class Method:
  BEGIN = 'calling.begin'
  ANSWER = 'calling.answer'
  END = 'calling.end'
  CONNECT = 'calling.connect'
  DISCONNECT = 'calling.disconnect'
  PLAY = 'calling.play'
  RECORD = 'calling.record'
  RECEIVE_FAX = 'calling.receive_fax'
  SEND_FAX = 'calling.send_fax'
  SEND_DIGITS = 'calling.send_digits'
  TAP = 'calling.tap'
  DETECT = 'calling.detect'
  PLAY_AND_COLLECT = 'calling.play_and_collect'

class Notification:
  STATE = 'calling.call.state'
  CONNECT = 'calling.call.connect'
  RECORD = 'calling.call.record'
  PLAY = 'calling.call.play'
  COLLECT = 'calling.call.collect'
  RECEIVE = 'calling.call.receive'
  FAX = 'calling.call.fax'
  DETECT = 'calling.call.detect'
  TAP = 'calling.call.tap'
  SEND_DIGITS = 'calling.call.send_digits'

class CallState:
  ALL = ['created', 'ringing', 'answered', 'ending', 'ended']
  NONE = 'none'
  CREATED = 'created'
  RINGING = 'ringing'
  ANSWERED = 'answered'
  ENDING = 'ending'
  ENDED = 'ended'

class ConnectState:
  DISCONNECTED = 'disconnected'
  CONNECTING = 'connecting'
  CONNECTED = 'connected'
  FAILED = 'failed'

class DisconnectReason:
  ERROR = 'error'
  BUSY = 'busy'

class CallPlayState:
  PLAYING = 'playing'
  ERROR = 'error'
  FINISHED = 'finished'

class PromptState:
  ERROR = 'error'
  NO_INPUT = 'no_input'
  NO_MATCH = 'no_match'
  DIGIT = 'digit'
  SPEECH = 'speech'

class MediaType:
  AUDIO = 'audio'
  TTS = 'tts'
  SILENCE = 'silence'
  RINGTONE = 'ringtone'

class CallRecordState:
  RECORDING = 'recording'
  NO_INPUT = 'no_input'
  FINISHED = 'finished'

class RecordType:
  AUDIO = 'audio'

class CallFaxState:
  PAGE = 'page'
  ERROR = 'error'
  FINISHED = 'finished'

class CallSendDigitsState:
  FINISHED = 'finished'

class CallTapState:
  TAPPING = 'tapping'
  FINISHED = 'finished'

class TapType:
  AUDIO = 'audio'

class DetectType:
  FAX = 'fax'
  MACHINE = 'machine'
  DIGIT = 'digit'

class DetectState:
  ERROR = 'error'
  FINISHED = 'finished'
  CED = 'CED'
  CNG = 'CNG'
  MACHINE = 'MACHINE'
  HUMAN = 'HUMAN'
  UNKNOWN = 'UNKNOWN'
  READY = 'READY'
  NOT_READY = 'NOT_READY'



import logging
from signalwire.blade.messages.execute import Execute
from signalwire.blade.handler import trigger
from signalwire.relay import BaseRelay
from .message import Message
from .send_result import SendResult
from .constants import Notification, Method

class Messaging(BaseRelay):

  @property
  def service(self):
    return 'messaging'

  def notification_handler(self, notification):
    notification['params']['event_type'] = notification['event_type']
    message = Message(notification['params'])
    if notification['event_type'] == Notification.STATE:
      trigger(self.client.protocol, message, suffix=self.ctx_state_unique(message.context))
    elif notification['event_type'] == Notification.RECEIVE:
      trigger(self.client.protocol, message, suffix=self.ctx_receive_unique(message.context))

  async def send(self, *, from_number, to_number, context, body=None, media=None, tags=None):
    params = {
      'from_number': from_number,
      'to_number': to_number,
      'context': context
    }
    if body:
      params['body'] = body
    if media:
      params['media'] = media
    if tags:
      params['tags'] = tags

    message = Execute({
      'protocol': self.client.protocol,
      'method': Method.SEND,
      'params': params
    })
    try:
      response = await self.client.execute(message)
      logging.info(response['result']['message'])
      return SendResult(response['result'])
    except Exception as error:
      logging.error(f'Messaging send error: {str(error)}')
      return SendResult()



from . import BaseComponent
from ..constants import Method, Notification, CallState
from ...event import Event

class Hangup(BaseComponent):
  def __init__(self, call, reason: str):
    super().__init__(call)
    self.control_id = call.tag
    self.reason = reason

  @property
  def event_type(self):
    return Notification.STATE

  @property
  def method(self):
    return Method.END

  @property
  def payload(self):
    return {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'reason': self.reason
    }

  def notification_handler(self, params):
    self.state = params.get('call_state', None)
    if self.state is None:
      return

    if self.state in self._events_to_await:
      self.unregister()
      self.completed = True
      self.successful = self.state == CallState.ENDED
      self.event = Event(self.state, params)
      if 'end_reason' in params:
        self.reason = params['end_reason']
      if self.has_future():
        self._future.set_result(True)



import logging
from signalwire.blade.handler import trigger
from signalwire.relay import BaseRelay
from .call import Call
from .constants import Notification, DetectState, ConnectState

class Calling(BaseRelay):
  def __init__(self, client):
    super().__init__(client)
    self.calls = []

  @property
  def service(self):
    return 'calling'

  def notification_handler(self, notification):
    notification['params']['event_type'] = notification['event_type']
    if notification['event_type'] == Notification.STATE:
      self._on_state(notification['params'])
    elif notification['event_type'] == Notification.RECEIVE:
      self._on_receive(notification['params'])
    elif notification['event_type'] == Notification.CONNECT:
      self._on_connect(notification['params'])
    elif notification['event_type'] == Notification.PLAY:
      self._on_play(notification['params'])
    elif notification['event_type'] == Notification.COLLECT:
      self._on_collect(notification['params'])
    elif notification['event_type'] == Notification.RECORD:
      self._on_record(notification['params'])
    elif notification['event_type'] == Notification.FAX:
      self._on_fax(notification['params'])
    elif notification['event_type'] == Notification.SEND_DIGITS:
      self._on_send_digits(notification['params'])
    elif notification['event_type'] == Notification.TAP:
      self._on_tap(notification['params'])
    elif notification['event_type'] == Notification.DETECT:
      self._on_detect(notification['params'])

  def new_call(self, *, call_type='phone', from_number, to_number, timeout=None):
    call = Call(calling=self)
    call.call_type = call_type
    call.from_number = from_number
    call.to_number = to_number
    call.timeout = timeout
    return call

  async def dial(self, *, call_type='phone', from_number, to_number, timeout=None):
    call = Call(calling=self)
    call.call_type = call_type
    call.from_number = from_number
    call.to_number = to_number
    call.timeout = timeout
    return await call.dial()

  def add_call(self, call):
    self.calls.append(call)

  def remove_call(self, call):
    try:
      self.calls.remove(call)
    except ValueError:
      logging.warn('Call to remove not found')

  def _get_call_by_id(self, call_id):
    for call in self.calls:
      if call.id == call_id:
        return call
    return None

  def _get_call_by_tag(self, tag):
    for call in self.calls:
      if call.tag == tag:
        return call
    return None

  def _on_state(self, params):
    call = self._get_call_by_id(params['call_id'])
    tag = params.get('tag', None)
    if call is None and tag is not None:
      call = self._get_call_by_tag(tag)

    if call is not None:
      if call.id is None:
        call.id = params['call_id']
        call.node_id = params['node_id']
      trigger(Notification.STATE, params, suffix=call.tag) # Notify components listening on State and Tag
      call._state_changed(params)
    elif 'call_id' in params and 'peer' in params:
      call = Call(calling=self, **params)
    else:
      logging.error('Unknown call {0}'.format(params['call_id']))

  def _on_receive(self, params):
    call = Call(calling=self, **params)
    trigger(self.client.protocol, call, suffix=self.ctx_receive_unique(call.context))

  def _on_connect(self, params):
    call = self._get_call_by_id(params['call_id'])
    state = params['connect_state']
    if call is not None:
      try:
        if state == ConnectState.CONNECTED:
          call.peer = self._get_call_by_id(params['peer']['call_id'])
        else:
          call.peer = None
      except KeyError:
        pass
      trigger(Notification.CONNECT, params, suffix=call.tag) # Notify components listening on Connect and Tag
      trigger(call.tag, params, suffix='connect.stateChange')
      trigger(call.tag, params, suffix=f"connect.{state}")

  def _on_play(self, params):
    call = self._get_call_by_id(params['call_id'])
    if call is not None:
      trigger(Notification.PLAY, params, suffix=params['control_id']) # Notify components listening on Play and control_id
      trigger(call.tag, params, suffix='play.stateChange')
      trigger(call.tag, params, suffix=f"play.{params['state']}")

  def _on_collect(self, params):
    call = self._get_call_by_id(params['call_id'])
    if call is not None:
      trigger(Notification.COLLECT, params, suffix=params['control_id']) # Notify components listening on Collect and control_id
      trigger(call.tag, params, suffix='prompt')

  def _on_record(self, params):
    call = self._get_call_by_id(params['call_id'])
    if call is not None:
      trigger(Notification.RECORD, params, suffix=params['control_id']) # Notify components listening on Record and control_id
      trigger(call.tag, params, suffix='record.stateChange')
      trigger(call.tag, params, suffix=f"record.{params['state']}")

  def _on_fax(self, params):
    call = self._get_call_by_id(params['call_id'])
    if call is not None:
      trigger(Notification.FAX, params, suffix=params['control_id']) # Notify components listening on Fax and control_id
      trigger(call.tag, params, suffix='fax.stateChange')
      try:
        trigger(call.tag, params, suffix=f"fax.{params['fax']['type']}")
      except KeyError:
        pass

  def _on_send_digits(self, params):
    call = self._get_call_by_id(params['call_id'])
    if call is not None:
      trigger(Notification.SEND_DIGITS, params, suffix=params['control_id']) # Notify components listening on SendDigits and control_id
      trigger(call.tag, params, suffix='sendDigits.stateChange')
      trigger(call.tag, params, suffix=f"sendDigits.{params['state']}")

  def _on_tap(self, params):
    call = self._get_call_by_id(params['call_id'])
    if call is not None:
      trigger(Notification.TAP, params, suffix=params['control_id']) # Notify components listening on Tap and control_id
      trigger(call.tag, params, suffix='tap.stateChange')
      trigger(call.tag, params, suffix=f"tap.{params['state']}")

  def _on_detect(self, params):
    call = self._get_call_by_id(params['call_id'])
    if call is not None:
      trigger(Notification.DETECT, params, suffix=params['control_id']) # Notify components listening on Detect and control_id
      try:
        event = params['detect']['params']['event']
        suffix = event if event == DetectState.FINISHED or event == DetectState.ERROR else 'update'
        trigger(call.tag, params['detect'], suffix=f"detect.{suffix}")
      except KeyError:
        pass



from . import BaseComponent
from ..constants import Method, Notification, PromptState
from ..helpers import prepare_collect_params, prepare_media_list
from ...event import Event
from .decorators import stoppable, has_volume_control

@stoppable
@has_volume_control
class Prompt(BaseComponent):

  def __init__(self, call, prompt_type, play, **kwargs):
    super().__init__(call)
    self._collect = prepare_collect_params(prompt_type, kwargs)
    self.play = prepare_media_list(play)
    self.volume_value = kwargs.get('volume', None)
    self.prompt_type = prompt_type
    self.confidence = None
    self.input = None
    self.terminator = None

  @property
  def event_type(self):
    return Notification.COLLECT

  @property
  def method(self):
    return Method.PLAY_AND_COLLECT

  @property
  def payload(self):
    tmp = {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id,
      'play': self.play,
      'collect': self._collect,
    }
    if self.volume_value is not None:
      tmp['volume'] = float(self.volume_value)
    return tmp

  def notification_handler(self, params):
    self.completed = True
    self.unregister()
    result = params['result']
    if result['type'] == PromptState.SPEECH:
      self.state = 'successful'
      self.successful = True
      self.input = result['params']['text']
      self.confidence = result['params']['confidence']
    elif result['type'] == PromptState.DIGIT:
      self.state = 'successful'
      self.successful = True
      self.input = result['params']['digits']
      self.terminator = result['params']['terminator']
    else:
      self.state = result['type']
      self.successful = False
    self.prompt_type = result['type']
    self.event = Event(self.prompt_type, result)
    if self.has_future():
      self._future.set_result(True)



import asyncio

def safe_invoke_callback(callback, *args, **kwargs):
  if asyncio.iscoroutinefunction(callback):
    asyncio.create_task(callback(*args, **kwargs))
  else:
    callback(*args, **kwargs)



import logging
from signalwire.blade.helpers import safe_invoke_callback
from .client import Client
from .constants import Constants

class Consumer:
  def __init__(self, **kwargs):
    self.project = kwargs.pop('project', None)
    self.token = kwargs.pop('token', None)
    self.host = kwargs.pop('host', Constants.HOST)
    self.contexts = kwargs.pop('contexts', [])
    self.Connection = kwargs.pop('Connection', None)
    self.client = None

  def setup(self):
    pass

  def ready(self):
    pass

  def teardown(self):
    pass

  def on_incoming_call(self, call):
    pass

  def on_task(self, message):
    pass

  def on_incoming_message(self, message):
    pass

  def on_message_state_change(self, message):
    pass

  def run(self):
    self.setup()
    self.check_requirements()
    self.client = Client(**self._build_relay_client_params())
    self.client.on('ready', self._client_ready)
    self.client.connect()
    self.teardown()

  def check_requirements(self):
    if self.project is None or self.token is None or len(self.contexts) <= 0:
      raise Exception('project, token and contexts instance attributes are required!')

  def _build_relay_client_params(self):
    params = {
      'project': self.project,
      'token': self.token
    }
    if self.host is not None:
      params['host'] = self.host
    if self.Connection is not None:
      params['connection'] = self.Connection
    return params

  async def _client_ready(self, client):
    await self.client.calling.receive(self.contexts, self.on_incoming_call)
    await self.client.tasking.receive(self.contexts, self.on_task)
    await self.client.messaging.receive(self.contexts, self.on_incoming_message)
    await self.client.messaging.state_change(self.contexts, self.on_message_state_change)
    safe_invoke_callback(self.ready)



from unittest import TestCase
import os, pytest
from signalwire.rest import Client as signalwire_client
import vcr

my_vcr = vcr.VCR(
    cassette_library_dir='fixtures/cassettes',
    record_mode='once',
)

@pytest.fixture(scope="module")
def client():
  client = signalwire_client(os.getenv('SIGNALWIRE_ACCOUNT','signalwire-account-123'), os.getenv('SIGNALWIRE_TOKEN', '123456'), signalwire_space_url = os.getenv('SIGNALWIRE_SPACE', 'myaccount.signalwire.com'))
  return client

@my_vcr.use_cassette()
def test_accounts(client):
  account = client.api.accounts(os.getenv('SIGNALWIRE_ACCOUNT','signalwire-account-123')).fetch()
  assert(account.friendly_name == 'LAML testing') 
    

@my_vcr.use_cassette()
def test_applications(client):
  applications = client.applications.list()
  assert(applications[0].sid == '34f49a97-a863-4a11-8fef-bc399c6f0928')

@my_vcr.use_cassette()
def test_local_numbers(client):
  numbers = client.available_phone_numbers("US") \
                .local \
                .list(in_region="WA")
  assert(numbers[0].phone_number == '+12064015921')

@my_vcr.use_cassette()
def test_toll_free_numbers(client):
  numbers = client.available_phone_numbers("US") \
                .toll_free \
                .list(area_code="310")
  assert(numbers[0].phone_number == '+13103590741')

@my_vcr.use_cassette()
def test_conferences(client):
  conferences = client.conferences.list()

  assert(conferences[0].sid == 'a811cb2c-9e5a-415d-a951-701f8e884fb5')

@my_vcr.use_cassette()
def test_conference_members(client):
  participants = client.conferences('a811cb2c-9e5a-415d-a951-701f8e884fb5') \
                     .participants \
                     .list()

  assert(participants[0].call_sid == '7a520324-684d-435c-87c2-ea7975f371d0')

@my_vcr.use_cassette()
def test_incoming_phone_numbers(client):
  incoming_phone_numbers = client.incoming_phone_numbers.list()

  assert(incoming_phone_numbers[0].phone_number == '+18990000001')

@my_vcr.use_cassette()
def test_messages(client):
  message = client.messages.create(
      from_='+15059999999',
      to='+15058888888',
      media_url='https://www.google.com/images/branding/googlelogo/1x/googlelogo_color_272x92dp.png'
  )

  assert(message.sid == 'cbad786b-fdcd-4d2a-bcb2-fff9df045008')

@my_vcr.use_cassette()
def test_media(client):
  media = client.messages('0da01046-5cca-462f-bc50-adae4e1307e1').media.list()
  assert(media[0].sid == 'a1ee4484-99a4-4996-b7df-fd3ceef2e9ec')

@my_vcr.use_cassette()
def test_recordings(client):
  recordings = client.recordings.list()
  assert(recordings[0].call_sid == 'd411976d-d319-4fbd-923c-57c62b6f677a')

@my_vcr.use_cassette()
def test_transcriptions(client):
  transcriptions = client.transcriptions.list()
  assert(transcriptions[0].recording_sid == 'e4c78e17-c0e2-441d-b5dd-39a6dad496f8')

@my_vcr.use_cassette()
def test_queues(client):
  queues = client.queues.list()
  assert(queues[0].sid == '2fd1bc9b-2e1f-41ac-988f-06842700c10d')

@my_vcr.use_cassette()
def test_queue_members(client):
  members = client.queues('2fd1bc9b-2e1f-41ac-988f-06842700c10d').members.list()
  assert(members[0].call_sid == '24c0f807-2663-4080-acef-c0874f45274d')

@my_vcr.use_cassette()
def test_send_fax(client):
  fax = client.fax.faxes.create(
    from_='+15556677888',
    to='+15556677999',
    media_url='https://www.w3.org/WAI/ER/tests/xhtml/testfiles/resources/pdf/dummy.pdf'
  )
  assert(fax.sid == 'dd3e1ac4-50c9-4241-933a-5d4e9a2baf31')

@my_vcr.use_cassette()
def test_list_fax(client):
  faxes = client.fax.faxes.list()
  assert(faxes[0].sid == 'dd3e1ac4-50c9-4241-933a-5d4e9a2baf31')

@my_vcr.use_cassette()
def test_fetch_fax(client):
  fax = client.fax.faxes('831455c6-574e-4d8b-b6ee-2418140bf4cd').fetch()
  assert(fax.to == '+15556677999')
  assert(fax.media_url == 'https://s3.us-east-2.amazonaws.com/signalwire-assets/faxes/20190104162834-831455c6-574e-4d8b-b6ee-2418140bf4cd.tiff')

@my_vcr.use_cassette()
def test_fetch_fax_media(client):
  media = client.fax.faxes('831455c6-574e-4d8b-b6ee-2418140bf4cd').media.list()
  assert(media[0].sid == 'aff0684c-3445-49bc-802b-3a0a488139f5')

@my_vcr.use_cassette()
def test_fetch_fax_media_instance(client):
  media = client.fax.faxes('831455c6-574e-4d8b-b6ee-2418140bf4cd').media('aff0684c-3445-49bc-802b-3a0a488139f5').fetch()
  assert(media.url == '/api/laml/2010-04-01/Accounts/signalwire-account-123/Faxes/831455c6-574e-4d8b-b6ee-2418140bf4cd/Media/aff0684c-3445-49bc-802b-3a0a488139f5.json')



import logging
from signalwire.blade.messages.execute import Execute
from signalwire.blade.messages.subscription import Subscription

# TODO: move protocols/methods to a constants file

async def setup_protocol(client):
  setup_message = Execute({
    'protocol': 'signalwire',
    'method': 'setup',
    'params': {}
  })
  response = await client.execute(setup_message)
  protocol = response['result']['protocol']
  subscribe_message = Subscription({
    'command': 'add',
    'protocol': protocol,
    'channels': ['notifications']
  })
  response = await client.execute(subscribe_message)
  return protocol

async def receive_contexts(client, contexts):
  contexts = list(set(contexts) - set(client.contexts))
  if len(contexts) == 0:
    return True
  logging.info(f'Trying to receive contexts: {contexts}')
  message = Execute({
    'protocol': client.protocol,
    'method': 'signalwire.receive',
    'params': {
      'contexts': contexts
    }
  })
  response = await client.execute(message)
  logging.info(response['result']['message'])
  client.contexts = list(set(client.contexts + contexts))
  return response['result']



class Method:
  SEND = 'messaging.send'

class Notification:
  STATE = 'messaging.state'
  RECEIVE = 'messaging.receive'

class MessageState:
  QUEUED = 'queued'
  INITIATED = 'initiated'
  SENT = 'sent'
  DELIVERED = 'delivered'
  UNDELIVERED = 'undelivered'
  FAILED = 'failed'

class DisconnectReason:
  ERROR = 'error'
  BUSY = 'busy'



from . import BaseResult

class DetectResult(BaseResult):

  @property
  def detect_type(self):
    return self.component.detect_type

  @property
  def result(self):
    return self.component.result



from . import BaseComponent
from ..constants import Method, Notification, CallPlayState
from ..helpers import prepare_media_list
from ...event import Event
from .decorators import stoppable, pausable, resumable, has_volume_control

@stoppable
@pausable
@resumable
@has_volume_control
class Play(BaseComponent):

  def __init__(self, call, play, volume=0):
    super().__init__(call)
    self.play = prepare_media_list(play)
    self.volume_value = float(volume)

  @property
  def event_type(self):
    return Notification.PLAY

  @property
  def method(self):
    return Method.PLAY

  @property
  def payload(self):
    tmp = {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id,
      'play': self.play
    }
    if self.volume_value != 0:
      tmp['volume'] = self.volume_value
    return tmp

  def notification_handler(self, params):
    self.state = params.get('state', None)
    if self.state is None:
      return

    self.completed = self.state != CallPlayState.PLAYING
    if self.completed:
      self.unregister()
      self.successful = self.state == CallPlayState.FINISHED
      self.event = Event(self.state, params)
      if self.has_future():
        self._future.set_result(True)



from . import BaseResult

class PlayResult(BaseResult):
  pass

class PlayPauseResult:
  def __init__(self, successful):
    self.successful = successful

class PlayResumeResult:
  def __init__(self, successful):
    self.successful = successful

class PlayVolumeResult:
  def __init__(self, successful):
    self.successful = successful



from . import BaseAction
from ..results.prompt_result import PromptResult, PromptVolumeResult
from ..results.stop_result import StopResult

class PromptAction(BaseAction):

  @property
  def result(self):
    return PromptResult(self.component)

  async def stop(self):
    result = await self.component.stop()
    return StopResult(result)

  async def volume(self, value):
    result = await self.component.volume(value)
    return PromptVolumeResult(result)



import json
from base64 import b64encode
from urllib.parse import urlencode
from urllib.request import Request, urlopen
from urllib.error import HTTPError
from .constants import Constants

class Task:
  def __init__(self, project, token, host=Constants.HOST):
    self.project = project
    self.token = token
    self.host = host

  def _authorization(self):
    data = f'{self.project}:{self.token}'.encode('utf-8')
    return 'Basic ' + str(b64encode(data), 'utf-8')

  def deliver(self, context, message):
    uri = f"https://{self.host}/api/relay/rest/tasks"
    data = json.dumps({ 'context': context, 'message': message }).encode('utf8')
    headers = {
      'Authorization': self._authorization(),
      'Content-Type': 'application/json',
      'Content-Length': len(data)
    }
    req = Request(uri, data=data, headers=headers)
    try:
      response = urlopen(req)
      return response.getcode() == 204
    except HTTPError as error:
      print('Task deliver error: {0}'.format(str(error)))
      return False



from . import BaseComponent
from ..constants import Method, Notification, DetectState, DetectType
from ...event import Event
from .decorators import stoppable

@stoppable
class Detect(BaseComponent):

  def __init__(self, call, detect_type, wait_for_beep=False, timeout=None, initial_timeout=None, end_silence_timeout=None, machine_voice_threshold=None, machine_words_threshold=None, tone=None, digits=None):
    super().__init__(call)
    self.detect_type = detect_type
    params = {}
    if initial_timeout is not None:
      params['initial_timeout'] = initial_timeout
    if end_silence_timeout is not None:
      params['end_silence_timeout'] = end_silence_timeout
    if machine_voice_threshold is not None:
      params['machine_voice_threshold'] = machine_voice_threshold
    if machine_words_threshold is not None:
      params['machine_words_threshold'] = machine_words_threshold
    if tone is not None:
      params['tone'] = tone
    if digits is not None:
      params['digits'] = digits
    self.detect = {
      'type': self.detect_type,
      'params': params
    }
    self.timeout = timeout
    self._wait_for_beep = wait_for_beep
    self._waiting_for_ready = False
    self.result = None
    self._results = []

  @property
  def event_type(self):
    return Notification.DETECT

  @property
  def method(self):
    return Method.DETECT

  @property
  def payload(self):
    tmp = {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id,
      'detect': self.detect
    }
    if self.timeout:
      tmp['timeout'] = self.timeout
    return tmp

  def notification_handler(self, params):
    try:
      detect = params['detect']
      self.detect_type = detect['type']
      self.state = detect['params']['event']
    except KeyError:
      return

    if self.state in [DetectState.FINISHED, DetectState.ERROR]:
      return self._complete(detect)

    if not self.has_future():
      self._results.append(self.state)
      return

    if self.detect_type == DetectType.DIGIT:
      return self._complete(detect)

    if self._waiting_for_ready:
      if self.state == DetectState.READY:
        return self._complete(detect)
      return

    if self._wait_for_beep and self.state == DetectState.MACHINE:
      self._waiting_for_ready = True
      return

    if self.state in self._events_to_await:
      return self._complete(detect)

  def _complete(self, detect):
    self.unregister()
    self.completed = True
    self.event = Event(self.state, detect)
    if self.has_future():
      # force READY/NOT_READY to MACHINE
      self.result = DetectState.MACHINE if self.state in [DetectState.READY, DetectState.NOT_READY] else self.state
      self.successful = self.state not in [DetectState.FINISHED, DetectState.ERROR]
      self._future.set_result(True)
    else:
      self.result = ','.join(self._results)
      self.successful = self.state != DetectState.ERROR



from abc import ABC
from ..components import BaseComponent

class BaseAction(ABC):
  def __init__(self, component: BaseComponent):
    self.component = component

  @property
  def control_id(self):
    return self.component.control_id

  @property
  def payload(self):
    return self.component.payload

  @property
  def completed(self):
    return self.component.completed

  @property
  def state(self):
    return self.component.state



from . import BaseResult

class FaxResult(BaseResult):

  @property
  def direction(self):
    return self.component.direction

  @property
  def identity(self):
    return self.component.identity

  @property
  def remote_identity(self):
    return self.component.remote_identity

  @property
  def document(self):
    return self.component.document

  @property
  def pages(self):
    return self.component.pages



from signalwire.blade.messages.execute import Execute

async def _execute(self, method):
  msg = Execute({
    'protocol': self.call.calling.client.protocol,
    'method': method,
    'params': {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id
    }
  })
  try:
    await self.call.calling.client.execute(msg)
    return True
  except Exception:
    return False

def stoppable(cls):
  def stop(self):
    return _execute(self, f'{self.method}.stop')
  setattr(cls, 'stop', stop)
  return cls

def pausable(cls):
  def pause(self):
    return _execute(self, f'{self.method}.pause')
  setattr(cls, 'pause', pause)
  return cls

def resumable(cls):
  def resume(self):
    return _execute(self, f'{self.method}.resume')
  setattr(cls, 'resume', resume)
  return cls

def has_volume_control(cls):
  async def volume(self, volume):
    msg = Execute({
      'protocol': self.call.calling.client.protocol,
      'method': f'{self.method}.volume',
      'params': {
        'node_id': self.call.node_id,
        'call_id': self.call.id,
        'control_id': self.control_id,
        'volume': float(volume)
      }
    })
    try:
      await self.call.calling.client.execute(msg)
      return True
    except Exception:
      return False

  setattr(cls, 'volume', volume)
  return cls



from . import BaseAction
from ..results.detect_result import DetectResult
from ..results.stop_result import StopResult

class DetectAction(BaseAction):

  @property
  def result(self):
    return DetectResult(self.component)

  async def stop(self):
    result = await self.component.stop()
    return StopResult(result)



from . import BaseComponent
from ..constants import Method, Notification, CallRecordState, RecordType
from ..helpers import prepare_record_params
from ...event import Event
from .decorators import stoppable

@stoppable
class Record(BaseComponent):

  def __init__(self, call, record_type=RecordType.AUDIO, beep=None, record_format=None, stereo=None, direction=None, initial_timeout=None, end_silence_timeout=None, terminators=None):
    super().__init__(call)
    self.url = ''
    self.duration = 0
    self.size = 0
    self._record = prepare_record_params(record_type, beep, record_format, stereo, direction, initial_timeout, end_silence_timeout, terminators)

  @property
  def event_type(self):
    return Notification.RECORD

  @property
  def method(self):
    return Method.RECORD

  @property
  def payload(self):
    return {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id,
      'record': self._record
    }

  def notification_handler(self, params):
    self.state = params.get('state', None)
    if self.state is None:
      return

    self.completed = self.state != CallRecordState.RECORDING
    if self.completed:
      self.unregister()
      self.successful = self.state == CallRecordState.FINISHED
      self.event = Event(self.state, params)
      self.url = params.get('url', '')
      self.duration = params.get('duration', 0)
      self.size = params.get('size', 0)
      if self.has_future():
        self._future.set_result(True)



from .base_fax import BaseFax
from ..constants import Method

class FaxReceive(BaseFax):

  @property
  def method(self):
    return Method.RECEIVE_FAX

  @property
  def payload(self):
    return {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id
    }



import base64
import hmac
import json
from hashlib import sha1
from urllib.parse import urlparse
from twilio.request_validator import compare, remove_port, add_port, RequestValidator as TwilioRequestValidator


class RequestValidator(object):
    def __init__(self, token):
        self.compatibility_validator = TwilioRequestValidator(token)
        self.token = token.encode()

    def build_signature_with_compatibility(self, uri, params):
        return self.compatibility_validator.compute_signature(uri, params)
    
    def validate_with_compatibility(self, uri, params, signature):
        return self.compatibility_validator.validate(uri, params, signature)
    
    def build_signature(self, uri, params):
        s = uri
        if params:
            s += params
        
        hmac_buffer = hmac.new(self.token, s.encode(), sha1)
        result = hmac_buffer.digest().hex()

        return result.strip()

    def validate(self, uri, params, signature):

        if isinstance(params, str):
            parsed_uri = urlparse(uri)
            with_port = add_port(parsed_uri)
            without_port = remove_port(parsed_uri)

            valid_signature_without_port = compare(
                self.build_signature(without_port, params),
                signature
            )
            valid_signature_with_port = compare(
                self.build_signature(with_port, params),
                signature
            )

            if valid_signature_without_port or valid_signature_with_port:
                return True

            try:
                parsed_params = json.loads(params)
                return self.validate_with_compatibility(uri, parsed_params, signature)
            except json.JSONDecodeError as e:
                return False
        
        return self.validate_with_compatibility(uri, params, signature)



from .base_fax import BaseFax
from ..constants import Method

class FaxSend(BaseFax):

  def __init__(self, call, document, identity=None, header=None):
    super().__init__(call)
    self._document = document
    self._identity = identity
    self._header = header

  @property
  def method(self):
    return Method.SEND_FAX

  @property
  def payload(self):
    tmp = {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id,
      'document': self._document
    }
    if self._identity:
      tmp['identity'] = self._identity
    if self._header:
      tmp['header_info'] = self._header
    return tmp



from signalwire.relay.calling.components import BaseComponent
from ..helpers import prepare_connect_devices, prepare_media_list
from ..constants import Method, Notification, ConnectState
from ...event import Event

class Connect(BaseComponent):
  def __init__(self, call, devices, ringback=[]):
    super().__init__(call)
    self.control_id = call.tag
    self.devices = prepare_connect_devices(devices, call.from_number, call.timeout)
    self.ringback = prepare_media_list(ringback)

  @property
  def event_type(self):
    return Notification.CONNECT

  @property
  def method(self):
    return Method.CONNECT

  @property
  def payload(self):
    tmp = {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'devices': self.devices
    }
    if len(self.ringback) > 0:
      tmp['ringback'] = self.ringback
    return tmp

  def notification_handler(self, params):
    self.state = params.get('connect_state', None)
    if self.state is None:
      return
    self.completed = self.state != ConnectState.CONNECTING
    if self.completed:
      self.unregister()
      self.successful = self.state == ConnectState.CONNECTED
      self.event = Event(self.state, params)
      if self.has_future():
        self._future.set_result(True)



import asyncio
from unittest.mock import MagicMock
from signalwire.blade.messages.message import Message

def AsyncMock(*args, **kwargs):
  m = MagicMock(*args, **kwargs)

  async def mock_coro(*args, **kwargs):
    return m(*args, **kwargs)

  mock_coro.mock = m
  return mock_coro

class MockedConnection:
  def __init__(self, client):
    self.client = client
    self.connect = AsyncMock()
    self.queue = asyncio.Queue()
    self.responses = []
    self.close = AsyncMock()
    self.connected = True

  async def send(self, message):
    await self.queue.put(message.id)

  async def read(self):
    for response in self.responses:
      msg = Message.from_json(response)
      msg.id = await self.queue.get()
      self.client.message_handler(msg)



from . import BaseAction
from ..results.play_result import PlayResult, PlayPauseResult, PlayResumeResult, PlayVolumeResult
from ..results.stop_result import StopResult

class PlayAction(BaseAction):

  @property
  def result(self):
    return PlayResult(self.component)

  async def stop(self):
    result = await self.component.stop()
    return StopResult(result)

  async def pause(self):
    result = await self.component.pause()
    return PlayPauseResult(result)

  async def resume(self):
    result = await self.component.resume()
    return PlayResumeResult(result)

  async def volume(self, value):
    result = await self.component.volume(value)
    return PlayVolumeResult(result)



from . import BaseResult

class RecordResult(BaseResult):

  @property
  def url(self):
    return self.component.url

  @property
  def duration(self):
    return self.component.duration

  @property
  def size(self):
    return self.component.size



from unittest import TestCase

class TestMessagingResponse(TestCase):
  def test_returns_laml(self):
    from signalwire.messaging_response import MessagingResponse
    r = MessagingResponse()
    r.message("This is a message from SignalWire!")
    self.assertEqual(str(r), '<?xml version="1.0" encoding="UTF-8"?><Response><Message>This is a message from SignalWire!</Message></Response>')




import os, codecs, re
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

def read(*file_paths):
  """Read file data."""
  with codecs.open(os.path.join(HERE, *file_paths), "r") as file_in:
    return file_in.read()

def get_version():
  content = read('signalwire/__init__.py')
  match = re.search(r"^__version__ = '(.*)'$", content, re.MULTILINE)
  if match:
    return match.group(1)
  raise RuntimeError('Unable to find package version!')

CLASSIFIERS = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Developers',
  'Intended Audience :: Information Technology',
  'Intended Audience :: Telecommunications Industry',
  'License :: OSI Approved :: MIT License',
  'Natural Language :: English',
  'Topic :: Communications',
  'Topic :: Software Development'
]

setup(
  name='signalwire',
  version=get_version(),
  description='Client library for connecting to SignalWire.',
  long_description=read('README.md'),
  long_description_content_type="text/markdown",
  classifiers=CLASSIFIERS,
  url='https://github.com/signalwire/signalwire-python',
  author='SignalWire Team',
  author_email='open.source@signalwire.com',
  license='MIT',
  packages=find_packages(exclude=['tests', 'tests.*']),
  install_requires=[
    'twilio==6.54.0',
    'aiohttp',
  ],
  python_requires='>=3.6',
  zip_safe=False
)



class Message:
  def __init__(self, params={}):
    self.id = params.get('message_id', None)
    self.state = params.get('message_state', None)
    self.context = params.get('context', None)
    self.from_number = params.get('from_number', None)
    self.to_number = params.get('to_number', None)
    self.body = params.get('body', None)
    self.direction = params.get('direction', None)
    self.media = params.get('media', None)
    self.segments = params.get('segments', None)
    self.tags = params.get('tags', None)
    self.reason = params.get('reason', None)



from signalwire import __version__
from signalwire.blade.messages.message import Message

class Connect(Message):

  MAJOR = 2
  MINOR = 3
  REVISION = 0

  def __init__(self, project, token):
    self.method = 'blade.connect'
    params = {
      'version': {
        'major': self.MAJOR,
        'minor': self.MINOR,
        'revision': self.REVISION
      },
      'authentication': {
        'project': project,
        'token': token
      },
      'agent': f'Python SDK/{__version__}'
    }
    super().__init__(params=params)



from . import BaseAction
from ..results.record_result import RecordResult
from ..results.stop_result import StopResult

class RecordAction(BaseAction):

  @property
  def result(self):
    return RecordResult(self.component)

  @property
  def url(self):
    return self.component.url

  async def stop(self):
    result = await self.component.stop()
    return StopResult(result)



class Constants:
  HOST = 'relay.signalwire.com'
  READY = 'ready'

class WebSocketEvents:
  OPEN = 'signalwire.socket.open'
  CLOSE = 'signalwire.socket.close'
  MESSAGE = 'signalwire.socket.message'
  ERROR = 'signalwire.socket.error'

class BladeMethod:
  NETCAST = 'blade.netcast'
  BROADCAST = 'blade.broadcast'
  DISCONNECT = 'blade.disconnect'



from . import BaseComponent
from ..constants import Method, Notification, CallTapState, TapType
from ...event import Event
from .decorators import stoppable

@stoppable
class Tap(BaseComponent):

  def __init__(self, call, audio_direction, target_type, target_addr=None, target_port=None, target_ptime=None, target_uri=None, rate=None, codec=None):
    super().__init__(call)
    self.tap = {
      'type': TapType.AUDIO,
      'params': { 'direction': audio_direction }
    }
    self.device = {
      'type': target_type,
      'params': {}
    }
    if target_addr is not None:
      self.device['params']['addr'] = target_addr
    if target_port is not None:
      self.device['params']['port'] = int(target_port)
    if target_ptime is not None:
      self.device['params']['ptime'] = int(target_ptime)
    if target_uri is not None:
      self.device['params']['uri'] = target_uri
    if rate is not None:
      self.device['params']['rate'] = int(rate)
    if codec is not None:
      self.device['params']['codec'] = codec

  @property
  def event_type(self):
    return Notification.TAP

  @property
  def method(self):
    return Method.TAP

  @property
  def payload(self):
    return {
      'node_id': self.call.node_id,
      'call_id': self.call.id,
      'control_id': self.control_id,
      'tap': self.tap,
      'device': self.device
    }

  @property
  def source_device(self):
    try:
      return self._execute_result['source_device']
    except Exception:
      return {}

  def notification_handler(self, params):
    self.state = params.get('state', None)
    if self.state is None:
      return
    if 'tap' in params:
      self.tap = params['tap']
    if 'device' in params:
      self.device = params['device']

    self.completed = self.state == CallTapState.FINISHED
    if self.completed:
      self.unregister()
      self.successful = True
      self.event = Event(self.state, params)
      if self.has_future():
        self._future.set_result(True)



from . import BaseAction
from ..results.tap_result import TapResult
from ..results.stop_result import StopResult

class TapAction(BaseAction):

  @property
  def result(self):
    return TapResult(self.component)

  @property
  def source_device(self):
    return self.component.source_device

  async def stop(self):
    result = await self.component.stop()
    return StopResult(result)



from unittest import TestCase
from multidict import MultiDict

class TestRequestValidator(TestCase):
    def test_should_validate_no_compatibility(self):
        from signalwire.request_validator import RequestValidator

        url = 'https://81f2-2-45-18-191.ngrok-free.app/'
        token = 'PSK_7TruNcSNTxp4zNrykMj4EPzF'
        signature = 'b18500437ebb010220ddd770cbe6fd531ea0ba0d'
        body = '{"call":{"call_id":"b5d63b2e-f75b-4dc8-b6d4-269b635f96c0","node_id":"fa3570ae-f8bd-42c2-83f4-9950d906c91b@us-west","segment_id":"b5d63b2e-f75b-4dc8-b6d4-269b635f96c0","call_state":"created","direction":"inbound","type":"phone","from":"+12135877632","to":"+12089806814","from_number":"+12135877632","to_number":"+12089806814","project_id":"4b7ae78a-d02e-4889-a63b-08b156d5916e","space_id":"62615f44-2a34-4235-b38b-76b5a1de6ef8"},"vars":{}}'

        validator = RequestValidator(token)
        computed = validator.build_signature(url, body)
        self.assertIsInstance(computed, str)
        self.assertEqual(signature, computed)
        valid = validator.validate(url, body, signature)
        self.assertTrue(valid)

    def test_should_validate_with_compatibity(self):
        from signalwire.request_validator import RequestValidator

        url = 'https://mycompany.com/myapp.php?foo=1&bar=2'
        token = '12345'
        signature = 'RSOYDt4T1cUTdK1PDd93/VVr8B8='
        body = {
            'CallSid': 'CA1234567890ABCDE',
            'Caller': '+14158675309',
            'Digits': '1234',
            'From': '+14158675309',
            'To': '+18005551212',
          }
        
        validator = RequestValidator(token)
        valid = validator.validate(url, body, signature)
        self.assertTrue(valid)

    def test_should_validate_with_compatibity_flask(self):
        from signalwire.request_validator import RequestValidator

        url = 'https://mycompany.com/myapp.php?foo=1&bar=2'
        token = '12345'
        signature = 'RSOYDt4T1cUTdK1PDd93/VVr8B8='
        body = MultiDict (
            [
                ('CallSid', 'CA1234567890ABCDE'),
                ('Caller', '+14158675309'),
                ('Digits', '1234'),
                ('From', '+14158675309'),
                ('To', '+18005551212')
            ]
          )
        
        validator = RequestValidator(token)
        valid = validator.validate(url, body, signature)
        self.assertTrue(valid)

    def test_should_validate_from_signalwire_http_request(self):
        from signalwire.request_validator import RequestValidator

        url = 'http://0aac-189-71-169-171.ngrok-free.app/voice'
        token = 'PSK_V3bF8oyeRNpJWGoRWHNYQMUU'
        signature = 'lf3nWPmUr2y6jSeeoMW4mg58vgI=' #From Lib
        body = {
            "AccountSid": "6bfbbe86-a901-4197-8759-2a0de1fa319d", 
            "ApiVersion": "2010-04-01", 
            "CallbackSource": "call-progress-events",  
            "CallSid": "0703574f-b151-465d-aedb-28972eb513c7",
            "CallStatus": "busy", 
            "Direction": "outbound-api", 
            "From": "sip:+17063958228@sip.swire.io",
            "HangupBy": "sip:jpsantos@joaosantos-2a0de1fa319d.sip.swire.io", 
            "HangupDirection": "inbound", 
            "Timestamp": "Thu, 09 Nov 2023 17:05:04 +0000", 
            "To": "sip:jpsantos@joaosantos-2a0de1fa319d.sip.swire.io", 
            "SipResultCode": "486"
        }
        
        validator = RequestValidator(token)
        valid = validator.validate(url, body, signature)
        self.assertTrue(valid)

    def test_should_validate_from_signalwire_https_request(self):
        from signalwire.request_validator import RequestValidator

        url = 'https://675d-189-71-169-171.ngrok-free.app/voice'
        token = 'PSK_V3bF8oyeRNpJWGoRWHNYQMUU'
        signature = 'muUMpldcBHlzuXGZ5gbw1ETZCYA='
        body = {
            "CallSid": "a97d4e8a-6047-4e2b-be48-fb96b33b5642", 
            "AccountSid": "6bfbbe86-a901-4197-8759-2a0de1fa319d", 
            "ApiVersion": "2010-04-01", 
            "Direction": "outbound-api", 
            "From": "sip:+17063958228@sip.swire.io", 
            "To": "sip:jpsantos@joaosantos-2a0de1fa319d.sip.swire.io", 
            "Timestamp": "Thu, 09 Nov 2023 14:40:55 +0000", 
            "CallStatus": "no-answer", 
            "CallbackSource": "call-progress-events", 
            "HangupDirection": "outbound", 
            "HangupBy": "sip:+17063958228@sip.swire.io", 
            "SipResultCode": "487"
        }

        
        validator = RequestValidator(token)
        valid = validator.validate(url, body, signature)
        self.assertTrue(valid)

    def test_should_validate_from_raw_json(self):
        from signalwire.request_validator import RequestValidator

        url = 'https://675d-189-71-169-171.ngrok-free.app/voice'
        token = 'PSK_V3bF8oyeRNpJWGoRWHNYQMUU'
        signature = 'muUMpldcBHlzuXGZ5gbw1ETZCYA='
        body = '''{
            "CallSid": "a97d4e8a-6047-4e2b-be48-fb96b33b5642", 
            "AccountSid": "6bfbbe86-a901-4197-8759-2a0de1fa319d", 
            "ApiVersion": "2010-04-01", 
            "Direction": "outbound-api", 
            "From": "sip:+17063958228@sip.swire.io", 
            "To": "sip:jpsantos@joaosantos-2a0de1fa319d.sip.swire.io", 
            "Timestamp": "Thu, 09 Nov 2023 14:40:55 +0000", 
            "CallStatus": "no-answer", 
            "CallbackSource": "call-progress-events", 
            "HangupDirection": "outbound", 
            "HangupBy": "sip:+17063958228@sip.swire.io", 
            "SipResultCode": "487"
        }'''

        
        validator = RequestValidator(token)
        valid = validator.validate(url, body, signature)
        self.assertTrue(valid)







from unittest import TestCase

class TestVoiceResponse(TestCase):
  def test_returns_laml(self):
    from signalwire.voice_response import VoiceResponse, Dial
    r = VoiceResponse()
    r.say("Welcome to SignalWire!")
    self.assertEqual(str(r), '<?xml version="1.0" encoding="UTF-8"?><Response><Say>Welcome to SignalWire!</Say></Response>')

  def test_supports_virtual_agent(self):
    from signalwire.voice_response import VoiceResponse
    r = VoiceResponse()
    connect = r.connect(action='http://example.com/action')
    connect.virtual_agent(connectorName='project', statusCallback='https://mycallbackurl.com')
    self.assertEqual(str(r), '<?xml version="1.0" encoding="UTF-8"?><Response><Connect action="http://example.com/action"><VirtualAgent connectorName="project" statusCallback="https://mycallbackurl.com" /></Connect></Response>')




import os
import asyncio
import logging
import signal
import aiohttp
from uuid import uuid4
from signalwire.blade.connection import Connection
from signalwire.blade.messages.connect import Connect
from signalwire.blade.messages.ping import Ping
from signalwire.blade.handler import register, unregister, trigger
from .helpers import setup_protocol
from .calling import Calling
from .tasking import Tasking
from .messaging import Messaging
from .message_handler import handle_inbound_message
from .constants import Constants, WebSocketEvents

class Client:
  PING_DELAY = 10

  def __init__(self, project, token, host=Constants.HOST, connection=Connection):
    self.loop = asyncio.get_event_loop()
    self.host = host
    self.project = project
    self.token = token
    self.attach_signals()
    self.connection = connection(self)
    self.uuid = str(uuid4())
    self._reconnect = True
    self.session_id = None
    self.signature = None
    self.protocol = None
    self.contexts = []
    self._calling = None
    self._tasking = None
    self._messaging = None
    self._requests = {}
    self._idle = False
    self._executeQueue = asyncio.Queue()
    self._pingInterval = None
    log_level = os.getenv('LOG_LEVEL', 'INFO').upper()
    logging.basicConfig(level=log_level)

  @property
  def connected(self):
    return self.connection.connected

  @property
  def calling(self):
    if self._calling is None:
      self._calling = Calling(self)
    return self._calling

  @property
  def tasking(self):
    if self._tasking is None:
      self._tasking = Tasking(self)
    return self._tasking

  @property
  def messaging(self):
    if self._messaging is None:
      self._messaging = Messaging(self)
    return self._messaging

  async def execute(self, message):
    if message.id not in self._requests:
      self._requests[message.id] = self.loop.create_future()

    if self._idle == True or self.connected == False:
      await self._executeQueue.put(message)
    else:
      await self.connection.send(message)
    result = await self._requests[message.id]
    del self._requests[message.id]
    return result

  def connect(self):
    while self._reconnect:
      self.loop.run_until_complete(self._connect())

  async def _connect(self):
    try:
      await self.connection.connect()
      asyncio.create_task(self.on_socket_open())
      await self.connection.read()
      self.on_socket_close()
    except aiohttp.client_exceptions.ClientConnectorError as error:
      trigger(WebSocketEvents.ERROR, error, suffix=self.uuid)
      logging.warn(f"{self.host} seems down..")
    try:
      logging.info('Connection closed..')
      await asyncio.sleep(5)
    except asyncio.CancelledError:
      pass

  async def disconnect(self):
    logging.info('Disconnection..')
    self._idle = True
    self._reconnect = False
    await self.connection.close()
    await self.cancel_pending_tasks()
    logging.info(f"Bye bye!")
    self.loop.stop()

  def on(self, event, callback):
    register(event=event, callback=callback, suffix=self.uuid)
    return self

  def off(self, event, callback=None):
    unregister(event=event, callback=callback, suffix=self.uuid)
    return self

  async def cancel_pending_tasks(self):
    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    logging.info(f"Cancelling {len(tasks)} outstanding tasks..")
    [task.cancel() for task in tasks]
    await asyncio.gather(*tasks, return_exceptions=True)

  def handle_signal(self):
    asyncio.create_task(self.disconnect())

  def attach_signals(self):
    for s in ('SIGHUP', 'SIGTERM', 'SIGINT'):
      try:
        self.loop.add_signal_handler(getattr(signal, s), self.handle_signal)
      except:
        pass

  def on_socket_close(self):
    if self._pingInterval:
      self._pingInterval.cancel()
    self.contexts = []
    trigger(WebSocketEvents.CLOSE, suffix=self.uuid)

  async def on_socket_open(self):
    try:
      self._idle = False
      trigger(WebSocketEvents.OPEN, suffix=self.uuid)
      result = await self.execute(Connect(project=self.project, token=self.token))
      self.session_id = result['sessionid']
      self.signature = result['authorization']['signature']
      self.protocol = await setup_protocol(self)
      await self._clearExecuteQueue()
      self._pong = True
      self.keepalive()
      logging.info('Client connected!')
      trigger(Constants.READY, self, suffix=self.uuid)
    except Exception as error:
      logging.error('Client setup error: {0}'.format(str(error)))
      await self.connection.close()

  def keepalive(self):
    async def send_ping():
      if self._pong is False:
        return await self.connection.close()
      self._pong = False
      await self.execute(Ping())
      self._pong = True
    asyncio.create_task(send_ping())
    self._pingInterval = self.loop.call_later(self.PING_DELAY, self.keepalive)

  async def _clearExecuteQueue(self):
    while True:
      if self._executeQueue.empty():
        break
      message = self._executeQueue.get_nowait()
      asyncio.create_task(self.connection.send(message))

  def message_handler(self, msg):
    trigger(WebSocketEvents.MESSAGE, msg, suffix=self.uuid)
    if msg.id not in self._requests:
      return handle_inbound_message(self, msg)

    if hasattr(msg, 'error'):
      self._set_exception(msg.id, msg.error)
    elif hasattr(msg, 'result'):
      try:
        if msg.result['result']['code'] == '200':
          self._set_result(msg.id, msg.result)
        else:
          self._set_exception(msg.id, msg.result['result'])
      except KeyError: # not a Relay with "result.result.code"
        self._set_result(msg.id, msg.result)

  def _set_result(self, uuid, result):
    self._requests[uuid].set_result(result)

  def _set_exception(self, uuid, error):
    # TODO: replace with a custom exception
    self._requests[uuid].set_exception(Exception(error['message']))



def prepare_connect_devices(devices, default_from, default_timeout=None, nested=False):
  final = []
  for device in devices:
    if isinstance(device, list):
      final.append(prepare_connect_devices(device, default_from, default_timeout, True))
    elif isinstance(device, dict):
      params = {
        'from_number': device.get('from_number', default_from),
        'to_number': device.get('to_number', '')
      }
      timeout = device.get('timeout', default_timeout)
      if timeout:
        params['timeout'] = int(timeout)
      tmp = {
        'type': device.get('call_type', 'phone'),
        'params': params
      }
      final.append(tmp if nested else [tmp])
  return final

def prepare_media_list(media_list):
  final = []
  for media in media_list:
    if not isinstance(media, dict):
      continue
    media_type = media.pop('type', '')
    final.append({
      'type': media_type,
      'params': media
    })
  return final

def prepare_record_params(record_type, beep, record_format, stereo, direction, initial_timeout, end_silence_timeout, terminators):
  params = {}
  if isinstance(beep, bool):
    params['beep'] = beep
  if record_format is not None:
    params['format'] = record_format
  if isinstance(stereo, bool):
    params['stereo'] = stereo
  if direction is not None:
    params['direction'] = direction
  if initial_timeout is not None:
    params['initial_timeout'] = initial_timeout
  if end_silence_timeout is not None:
    params['end_silence_timeout'] = end_silence_timeout
  if terminators is not None:
    params['terminators'] = terminators
  return { record_type: params }

def prepare_collect_params(prompt_type, params):
  collect = {}
  if prompt_type == 'both':
    collect['speech'] = {}
    collect['digits'] = {}
  elif prompt_type == 'digits':
    collect['digits'] = {}
  elif prompt_type == 'speech':
    collect['speech'] = {}

  # TODO: support partial_results
  if 'initial_timeout' in params:
    collect['initial_timeout'] = params['initial_timeout']
  if 'digits_max' in params:
    collect['digits']['max'] = params['digits_max']
  if 'digits_terminators' in params:
    collect['digits']['terminators'] = params['digits_terminators']
  if 'digits_timeout' in params:
    collect['digits']['digit_timeout'] = params['digits_timeout']
  if 'end_silence_timeout' in params:
    collect['speech']['end_silence_timeout'] = params['end_silence_timeout']
  if 'speech_timeout' in params:
    collect['speech']['speech_timeout'] = params['speech_timeout']
  if 'speech_language' in params:
    collect['speech']['language'] = params['speech_language']
  if 'speech_hints' in params:
    collect['speech']['hints'] = params['speech_hints']

  return collect

def prepare_prompt_media_list(params, kwargs):
  # helper method to build media_list for prompt_ringtone and prompt_tts
  for k in ['duration', 'language', 'gender']:
    if k in kwargs:
      params[k] = kwargs[k]
  return params
